/*
 * Name: Solar Tracker EspCode.ino (test version)
 * Author: Sorin Liviu Jurj
 * 
 * Description:
 * This software will implement different testing testing techniques to test 
 * the software functionality and features. We will tackle each part of the 
 * code individually and attempt to show all the possible break points and 
 * try to detect possible fault errors using white box testing techniques.
 *    This testing is based on the AUnit arduino library which is a port of
 * ArduinoUnit and Google Test testing platforms.
 */



#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <ArduinoJson.h> 
#include <SoftwareSerial.h>
#include <ctype.h>

// Library for testing 
#include <AUnit.h>
using aunit::TestRunner;      // name space for aunit to prevent name clashing 

//WiFi Variables
const char* ssid = "Home Network";
const char* password = "YOUR WIFI PASSWORD HERE";
WiFiClient espClient;

// MQTT Variables
String chipID = " ";

/*
 * must be current system iP address
 */
const char* mqtt_server = "159.69.5.173";


PubSubClient client(espClient);
char  InTopic[16] = "in";
char  OutTopic[16] = "Out";
char  chipID_c[16] = "in";

// Booleans
bool _Start = false;

volatile int ltsensor = 0; 
volatile int rtsensor = 0;  
volatile int rdsensor = 0; 
volatile int ldsensor = 0; 

String gh = ""; 
 
// 14 and 12 are pin D5, and D6 ON THE NODEMCI ESP12
SoftwareSerial mySerial(14, 12, false, 256);


//---------------------------------------------
//---------------------------------------------
//                TESTS
//---------------------------------------------
//--------------------------------------------- 
/*
 * This test is configured to run only once and
 * ensures that our ssid string contains only valid characters.
 */
test(ssidTest)
{
  for(uint8_t i=0; i<sizeof(ssid); i++)
  {
    assertMoreOrEqual(ssid[i], ' ');
  }
}


/*
 * This test is configured to run only once and
 * ensures that our password string contains only valid characters.
 */
test(passwordTest)
{
  for(uint8_t i=0; i<sizeof(password); i++)
  {
    assertMoreOrEqual(password[i], ' ');
  }
}


/*
 * This test is configured to run only once and
 * ensures that our mqtt address string is in the right format, 
 */
test(mqttAddrTest)
{
  for(uint8_t i=0; i<sizeof(mqtt_server); i++)
  {
    assertMoreOrEqual(mqtt_server[i], '.');
    assertLessOrEqual(mqtt_server[i], '9');
  }
}


/*
 * This test a test to check the inTopic array
 * which is dynamically generated using strings
 * which are notorious for memory problems. In our 
 * test we check that all the characters are valid 
 * characters and also that the length of the array is 
 * less than 10
 */
test(InTopicTest)
{
  uint8_t b = sizeof(InTopic);
  assertLess(b, 10);    // assert that the size of the array is less than 10
  
  for(uint8_t i=0; i<sizeof(InTopic); i++)
  {
    assertMoreOrEqual(InTopic[i], ' ');
  }
}


/*
 * This test checks if the value of the sensor data received is in the
 * reasonable range. If this value is outside the range 0 - 1024, we can
 * concluse that the data was corrupted during transmission or a memmory 
 * problem has occured.
 * This is a continous test and will be run over and over again.
 */
testing(sensorDataTest)
{
  assertEqual(isDigit(ltsensor), 1);      // makes sure that its a digit and not NAN
  assertEqual(isDigit(rtsensor), 1);
  assertEqual(isDigit(rdsensor), 1);
  assertEqual(isDigit(ldsensor), 1);

  assertMoreOrEqual(ltsensor, 0);         // makes sure it is an unsigned integer, i.e greater than zero
  assertMoreOrEqual(rtsensor, 0);
  assertMoreOrEqual(rdsensor, 0);
  assertMoreOrEqual(ldsensor, 0);

  assertLessOrEqual(ltsensor, 1023);      // makes sure that it is less than 1023
  assertLessOrEqual(rtsensor, 1023);
  assertLessOrEqual(rdsensor, 1023);
  assertLessOrEqual(ldsensor, 1023);
}


/*
 * We also have a test which tests the analog data received from the microcontroller 
 * to make sure that it is within the specified range.
 */
testing(incominingData)
{
   uint8_t a = gh.length()-1;
   assertLessOrEqual(a, 13); 
}


/*
 * Main Setup
 */
void setup()
{
  Serial.begin(9600);
  mySerial.begin(9600);
  
  /*
   * The while statement is used to ensure that the serial communication is completely
   * started before continuing with the program. This will prevent framing or data
   * corruption errors if we try to send or receive data when the serial comm is not
   * completely set up.
   */
  while(!Serial && !mySerial);

  /*
   * It is advisable to disable the watchdog timer, on the ESP. The main duty of the watch dog timer
   * it to reset the device if an error is detected (for example: if the device gets stuck in an
   * infinite loop and no instructions are being processed). It is very important in embedded software 
   * development as it manages unpredicatable devices for us by reseting the system, which gives the system
   * a chance to fix the error.
   */
  ESP.wdtDisable();

  
  delay(1000);
  
  // Wifi connetion
  Wifi_Connect();
  delay(500);
  
  MQTT_Connect();
}





///////////////////////////////////////////////
/////////          MAIN LOOP          ////////
/////////////////////////////////////////////
long lastTime1 = 0;
void loop()
{
  /*
   * We have to feed the Watch Dog timer regularly so that it does not timeout or
   * think that it has encountered an error. This is a precautionary measure as the 
   * watchdog timer normally feeds itself when instruction are being processed.
   * But it is a good practice to feed the watchdog timer at regular intervals just to 
   * account for unseen errors or delays in our code.
   */
  ESP.wdtFeed();
  Receive_Data();    // receive data 

  /*
   * Sometimes connection to the MQTT Server can break for many reasons. conditional 
   * statement will attempt to reconnect to the server if the connectin is broker.
   * But if the reconnection fails, we then restart the system.
   */
  if (!client.connected()) 
  {
     MQTT_Connect();

     // we add a 1 second delay to the reconnect loop so that we do not bang the server
     // with too many requests.
     delay(1000);
  }

  if(millis()-lastTime1 > 3000)
  {
    //client.publish(OutTopic, "Idle");
    lastTime1 = millis();
  }
  
  delay(100);
  client.loop();

  // this command runs all the defined tests.
  TestRunner::run();      // run test from test queue
  
}



///////////////////////////////////////////////
/////     MQTT AND WIFI FUNCTIONS        //////
/////////////////////////////////////////////
void Wifi_Connect() 
{
  /*
   * Before we try to connect we have to run a test which makes sure that the ssid 
   * and password for the wifi connection contain only valid characters.
   */
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);

  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
    ESP.wdtFeed();
  }

  // Generate ID
  chipID =  (String)ESP.getChipId();
  chipID.toCharArray(chipID_c, chipID.length()+1);
  Serial.print("Device ID"); Serial.println(chipID_c);
  
  // Generate Output Topic
  String s_out = "From-";
  s_out += chipID;
  s_out.toCharArray(OutTopic, s_out.length()+1);

  // Generate Input Topic
  String s_in = "To-";
  s_in += chipID;
  s_in.toCharArray(InTopic, s_in.length()+1);
  
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  Serial.print("Output Topic: "); Serial.println(OutTopic);
  Serial.print("Input Topic: "); Serial.println(InTopic);
}


void MQTT_Connect()
{
  client.setServer(mqtt_server, 1883);
  client.setCallback(MQTT_Receive_Callback);

  Serial.print("In Topic: "); Serial.println(InTopic);
  Serial.print("Out Topic: "); Serial.println(OutTopic);

  delay(100);
  
  int retry_counter = 0;

  while (retry_counter < 20)        // try to connect to web server 20 times
  {
    ESP.wdtFeed();
    Serial.print("Attempting MQTT connection to Web Server");
    
    if (client.connect(chipID_c)) 
    {
      Serial.println("Connected to Web MQTT Server");
      //client.publish(OutTopic, "ESP connected");
      client.subscribe(InTopic);
      Serial.print("Successfully suscribed to topic ");
      Serial.println(InTopic);
      return ;
    }
    else {
      Serial.print("failed, rc=");
      Serial.println(client.state());
      delay(1000);
    }
    retry_counter++;
  }

  // Restart device if connections fail
  ESP.restart();
}



void MQTT_Receive_Callback(char* topic, byte* payload, unsigned int length) 
{
  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.print("] "); Serial.println("");
  for (int i = 0; i < length; i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();

  ltsensor = getValue((char *)payload, '.', 1).toInt();
  rtsensor = getValue((char *)payload, '.', 2).toInt();
  ldsensor = getValue((char *)payload, '.', 3).toInt();
  rdsensor = getValue((char *)payload, '.', 4).toInt();

  UNO_Send();

  // send acknowledgement message back to server
  //client.publish(OutTopic, "Command received");
}


void UNO_Send()
{
  ESP.wdtFeed();
  Serial.println("sending to uno");
  mySerial.print(ltsensor); 
  mySerial.print('.');
  mySerial.print(rtsensor); 
  mySerial.print('.');
  mySerial.print(ldsensor); 
  mySerial.print('.');
  mySerial.print(rdsensor); 
  mySerial.print('.');
    
  mySerial.print('\r');
}




/*
 * Function that receives data from the arduino uno.
 * We have to run tests to confirm the integrity of the received data.
 */
void Receive_Data()
{
  if(mySerial.available())
  {
    gh = ""; 
    gh = mySerial.readStringUntil('\r');
    char *outdata = " ";
    gh.toCharArray(outdata, gh.length()-1);
    
    Serial.println("Publishing data to mqtt");
    Serial.println(gh);
    client.publish(OutTopic, outdata);    
    return ;
  }
}


String getValue(String data, char separator, int index)
{
    int found = 0;
    int strIndex[] = { 0, -1 };
    int maxIndex = data.length() - 1;

    for (int i = 0; i <= maxIndex && found <= index; i++) {
        if (data.charAt(i) == separator || i == maxIndex) {
            found++;
            strIndex[0] = strIndex[1] + 1;
            strIndex[1] = (i == maxIndex) ? i+1 : i;
        }
    }
    return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}


